@@ -5,45 +5,46 @@ pub struct Section {
     pub name: String,
     pub caption: Option<String>,
     pub header: Header,
-    pub body: Option<String>,
+    pub body: Option<(usize, String)>,
     pub sub_sections: SubSections,
     pub is_commented: bool,
+    pub line_number: usize,
 }
 
 impl Section {
-    pub fn body_without_comment(&self) -> Option<String> {
+    pub fn body_without_comment(&self) -> Option<(usize, String)> {
         let body = match &self.body {
             None => return None,
             Some(b) => b,
         };
         match body {
-            _ if body.starts_with(r"\/") =>
+            _ if body.1.starts_with(r"\/") =>
             {
                 #[allow(clippy::single_char_pattern)]
-                Some(body.strip_prefix(r"\").expect("").to_string())
+                Some((body.0, body.1.strip_prefix(r"\").expect("").to_string()))
             }
-            _ if body.starts_with('/') => None,
-            _ => Some(body.to_string()),
+            _ if body.1.starts_with('/') => None,
+            _ => Some((body.0, body.1.to_string())),
         }
     }
 
     pub fn remove_comments(&self) -> Section {
         let mut headers = vec![];
-        for (k, v) in self.header.0.iter() {
+        for (i, k, v) in self.header.0.iter() {
             if !k.starts_with('/') {
-                headers.push((k.to_string(), v.to_string()));
+                headers.push((i.to_owned(), k.to_string(), v.to_string()));
             }
         }
 
         let body = match &self.body {
             None => None,
             Some(body) => match body {
-                _ if body.starts_with(r"\/") =>
+                _ if body.1.starts_with(r"\/") =>
                 {
                     #[allow(clippy::single_char_pattern)]
-                    Some(body.strip_prefix(r"\").expect("").to_string())
+                    Some((body.0, body.1.strip_prefix(r"\").expect("").to_string()))
                 }
-                _ if body.starts_with('/') => None,
+                _ if body.1.starts_with('/') => None,
                 _ => self.body.clone(),
             },
         };
@@ -62,34 +63,42 @@ impl Section {
                     .collect::<Vec<SubSection>>(),
             ),
             is_commented: false,
+            line_number: self.line_number,
         }
     }
 
-    pub fn caption(&self) -> Result<String> {
+    pub fn caption(&self, line_number: usize, doc_id: &str) -> Result<String> {
         match self.caption {
             Some(ref v) => Ok(v.to_string()),
-            None => Err(Error::InvalidInput {
-                message: "caption is missing".to_string(),
-                context: "".to_string(),
+            None => Err(Error::ParseError {
+                message: format!("caption is missing in {}", self.name.as_str(),),
+                doc_id: doc_id.to_string(),
+                line_number,
             }),
         }
     }
 
-    pub fn body(&self) -> Result<String> {
+    pub fn body(&self, line_number: usize, doc_id: &str) -> Result<String> {
         match self.body_without_comment() {
-            Some(ref v) => Ok(v.to_string()),
-            None => Err(Error::InvalidInput {
-                message: "body is missing".to_string(),
-                context: "".to_string(),
+            Some(ref v) => Ok(v.1.to_string()),
+            None => Err(Error::ParseError {
+                message: format!("body is missing in {}", self.name.as_str(),),
+                doc_id: doc_id.to_string(),
+                line_number,
             }),
         }
     }
 
-    pub fn assert_missing(&self, key: &str) -> Result<()> {
-        if self.header.str_optional(key)?.is_some() {
-            return Err(Error::InvalidInput {
-                message: format!("'{}' is not expected", key),
-                context: "".to_string(),
+    pub fn assert_missing(&self, line_number: usize, key: &str, doc_id: &str) -> Result<()> {
+        if self
+            .header
+            .str_optional(doc_id.to_string(), line_number, key)?
+            .is_some()
+        {
+            return Err(Error::ParseError {
+                message: format!("'{}' is not expected in {}", key, self.name.as_str()),
+                doc_id: doc_id.to_string(),
+                line_number,
             });
         }
 
@@ -104,6 +113,19 @@ impl Section {
             body: None,
             sub_sections: SubSections::default(),
             is_commented: false,
+            line_number: 0,
+        }
+    }
+
+    pub fn without_line_number(&self) -> Self {
+        Self {
+            name: self.name.to_string(),
+            caption: self.caption.to_owned(),
+            header: self.header.without_line_number(),
+            body: self.body.to_owned().map(|v| (0, v.1)),
+            sub_sections: self.sub_sections.without_line_number(),
+            is_commented: self.is_commented.to_owned(),
+            line_number: 0,
         }
     }
 
@@ -120,7 +142,7 @@ impl Section {
     }
 
     pub fn add_header(mut self, key: &str, value: &str) -> Self {
-        self.header.0.push((key.to_string(), value.to_string()));
+        self.header.0.push((0, key.to_string(), value.to_string()));
         self
     }
 
@@ -156,12 +178,12 @@ impl Section {
     }
 
     pub fn and_body(mut self, body: &str) -> Self {
-        self.body = Some(body.to_string());
+        self.body = Some((0, body.to_string()));
         self
     }
 
     pub fn and_optional_body(mut self, body: &Option<String>) -> Self {
-        self.body = body.as_ref().map(|v| v.to_string());
+        self.body = body.as_ref().map(|v| (0, v.to_string()));
         self
     }
 
@@ -170,7 +192,11 @@ impl Section {
         self
     }
 
-    pub fn sub_section_by_name(&self, name: &str) -> crate::p1::Result<&crate::p1::SubSection> {
+    pub fn sub_section_by_name(
+        &self,
+        name: &str,
+        doc_id: String,
+    ) -> crate::p1::Result<&crate::p1::SubSection> {
         let mut count = 0;
         for s in self.sub_sections.0.iter() {
             if s.is_commented {
@@ -183,6 +209,8 @@ impl Section {
         if count > 1 {
             return Err(crate::p1::Error::MoreThanOneSubSections {
                 key: name.to_string(),
+                doc_id,
+                line_number: self.line_number,
             });
         }
 
@@ -196,6 +224,8 @@ impl Section {
         }
 
         Err(crate::p1::Error::NotFound {
+            doc_id,
+            line_number: self.line_number,
             key: name.to_string(),
         })
     }
