@@ -22,6 +22,7 @@ impl Document {
                 let value = match value {
                     ftd::Value::Boolean { value } => value.to_string(),
                     ftd::Value::Integer { value } => value.to_string(),
+                    ftd::Value::String { text: value, .. } => value.to_string(),
                     _ => continue,
                 };
                 d.insert(k.to_string(), value);
@@ -52,11 +53,12 @@ impl Document {
         }
         ftd_rt::Element::get_visible_event_dependencies(&self.main.container.children, &mut data);
         ftd_rt::Element::get_value_event_dependencies(&self.main.container.children, &mut data);
+        ftd_rt::Element::get_style_event_dependencies(&self.main.container.children, &mut data);
 
         data
     }
 
-    pub fn rerender(&mut self, id: &str) -> crate::p1::Result<ftd_rt::Document> {
+    pub fn rerender(&mut self, id: &str, doc_id: &str) -> crate::p1::Result<ftd_rt::Document> {
         let mut rt = ftd::RT::from(
             self.name.as_str(),
             self.aliases.clone(),
@@ -67,27 +69,27 @@ impl Document {
         let data = self.rt_data();
         Ok(ftd_rt::Document {
             data,
-            html: self.html(id),
+            html: self.html(id, doc_id),
             external_children: ftd_rt::Element::get_external_children_dependencies(
                 &self.main.container.children,
             ),
         })
     }
 
-    pub fn to_rt(&self, id: &str) -> ftd_rt::Document {
+    pub fn to_rt(&self, id: &str, doc_id: &str) -> ftd_rt::Document {
         let external_children =
             ftd_rt::Element::get_external_children_dependencies(&self.main.container.children);
 
         ftd_rt::Document {
             data: self.rt_data(),
-            html: self.html(id),
+            html: self.html(id, doc_id),
             external_children,
         }
     }
 
-    pub fn html(&self, id: &str) -> String {
+    pub fn html(&self, id: &str, doc_id: &str) -> String {
         self.main
-            .to_node()
+            .to_node(doc_id)
             .to_html(&Default::default(), &self.rt_data(), id)
     }
 
@@ -137,6 +139,16 @@ impl Document {
                             return Some(v);
                         }
                     }
+                    ftd_rt::Element::TextBlock(_) => {
+                        if let Some(v) = f(e) {
+                            return Some(v);
+                        }
+                    }
+                    ftd_rt::Element::Code(_) => {
+                        if let Some(v) = f(e) {
+                            return Some(v);
+                        }
+                    }
                     ftd_rt::Element::Input(_) => {
                         if let Some(v) = f(e) {
                             return Some(v);
@@ -160,6 +172,15 @@ impl Document {
                             return Some(t);
                         }
                     }
+                    ftd_rt::Element::Scene(c) => {
+                        if let Some(v) = f(e) {
+                            return Some(v);
+                        }
+
+                        if let Some(t) = finder(&c.container.children, f) {
+                            return Some(t);
+                        }
+                    }
                     ftd_rt::Element::Image(_) => {
                         if let Some(v) = f(e) {
                             return Some(v);
@@ -243,7 +264,7 @@ impl Document {
         Ok(d)
     }
 
-    fn get_heading<F>(children: &[ftd_rt::Element], f: &F) -> Option<ftd_rt::Rendered>
+    pub fn get_heading<F>(children: &[ftd_rt::Element], f: &F) -> Option<ftd_rt::Rendered>
     where
         F: Fn(&ftd_rt::Region) -> bool,
     {
@@ -341,11 +362,12 @@ impl Document {
             return Ok(None);
         }
         if v.len() > 1 {
-            return crate::e(format!(
-                "more than one instances({}) of {} found",
-                v.len(),
-                record
-            ));
+            return ftd::e2(
+                format!("more than one instances({}) of {} found", v.len(), record),
+                self.name.as_str(),
+                self.name.to_string(),
+                0,
+            );
         }
         Ok(Some(v.into_iter().next().unwrap())) // unwrap okay because v not empty
     }
@@ -371,7 +393,7 @@ impl Document {
                 }
                 serde_json::Value::Array(a)
             }
-            t => return ftd::e2("not a record", t),
+            t => return ftd::e2("not a record", t, self.name.to_string(), 0),
         };
 
         Ok(serde_json::from_value(json)?)
@@ -408,6 +430,8 @@ impl Document {
             Some(v) => v,
             None => {
                 return Err(crate::p1::Error::NotFound {
+                    doc_id: "".to_string(),
+                    line_number: 0,
                     key: key.to_string(),
                 })
             }
@@ -436,7 +460,7 @@ impl Document {
             } => self.object_to_json(Some(variant), fields)?,
             crate::Value::List { data, .. } => self.list_to_json(data)?,
             crate::Value::None { .. } => serde_json::Value::Null,
-            _ => return ftd::e2("unhandled value found(value_to_json)", v),
+            _ => return ftd::e2("unhandled value found(value_to_json)", v, "".to_string(), 0),
         })
     }
 
@@ -487,6 +511,82 @@ impl Document {
     }
 }
 
+pub fn set_region_id(elements: &mut Vec<ftd_rt::Element>) {
+    let mut map: std::collections::BTreeMap<usize, String> = Default::default();
+    for element in elements.iter_mut() {
+        match element {
+            ftd_rt::Element::Column(ftd_rt::Column { container, .. })
+            | ftd_rt::Element::Row(ftd_rt::Row { container, .. }) => {
+                set_region_id(&mut container.children);
+                if let Some((_, _, ref mut e)) = container.external_children {
+                    set_region_id(e);
+                }
+            }
+            _ => continue,
+        }
+    }
+
+    for (idx, element) in elements.iter().enumerate() {
+        match element {
+            ftd_rt::Element::Column(ftd_rt::Column { common, .. })
+            | ftd_rt::Element::Row(ftd_rt::Row { common, .. }) => {
+                if common.region.as_ref().filter(|v| v.is_heading()).is_some()
+                    && common.data_id.is_none()
+                {
+                    if let Some(h) =
+                        ftd::p2::Document::get_heading(vec![element.clone()].as_slice(), &|r| {
+                            r.is_heading()
+                        })
+                    {
+                        map.insert(idx, slug::slugify(h.original));
+                    }
+                }
+            }
+            _ => continue,
+        }
+    }
+    for (idx, s) in map {
+        elements[idx].get_mut_common().unwrap().id = Some(s);
+    }
+}
+
+pub fn default_scene_children_position(elements: &mut Vec<ftd_rt::Element>) {
+    for element in elements {
+        if let ftd_rt::Element::Scene(scene) = element {
+            for child in &mut scene.container.children {
+                check_and_set_default_position(child);
+            }
+            if let Some((_, _, ref mut ext_children)) = scene.container.external_children {
+                for child in ext_children {
+                    check_and_set_default_position(child);
+                }
+            }
+        }
+        match element {
+            ftd_rt::Element::Scene(ftd_rt::Scene { container, .. })
+            | ftd_rt::Element::Row(ftd_rt::Row { container, .. })
+            | ftd_rt::Element::Column(ftd_rt::Column { container, .. }) => {
+                default_scene_children_position(&mut container.children);
+                if let Some((_, _, ref mut ext_children)) = container.external_children {
+                    default_scene_children_position(ext_children);
+                }
+            }
+            _ => {}
+        }
+    }
+
+    fn check_and_set_default_position(child: &mut ftd_rt::Element) {
+        if let Some(common) = child.get_mut_common() {
+            if common.top.is_none() && common.bottom.is_none() {
+                common.top = Some(0);
+            }
+            if common.left.is_none() && common.right.is_none() {
+                common.left = Some(0);
+            }
+        }
+    }
+}
+
 #[cfg(test)]
 mod test {
     use crate::test::*;
@@ -498,7 +598,7 @@ mod test {
             indoc::indoc!(
                 "
             -- import: fifthtry/ft
-            -- ft.toc:
+            -- $ft.toc:
 
             foo is the toc
             "
@@ -541,13 +641,12 @@ mod test {
                 who: caption
 
                 -- record meta_type:
-                license: string
-                reader: list someone
+                string license:
+                someone list reader:
 
-                -- list meta:
-                type: meta_type
+                -- meta_type list $meta:
 
-                -- meta:
+                -- $meta:
                 license: BSD
 
                 --- reader.Username: foo
@@ -629,8 +728,8 @@ mod test {
             indoc::indoc!(
                 "
                 -- record pr:
-                number: integer
-                title: caption
+                integer number:
+                caption title:
 
                 -- pr: some pr
                 number: 24
